/*
 * Licensed Materials - Property of IBM
 * 
 * (C) Copyright IBM Corporation. 2012,2014 All Rights Reserved
 * 
 * US Government Users Restricted Rights - Use, duplication, or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corporation
 * 
 * 
 * The following code is sample code created by IBM Corporation. This sample
 * code is not part of any standard IBM product and is provided to you solely
 * for the purpose of assisting you in the development of your applications. The
 * code is provided 'AS IS', without warranty or condition of any kind. IBM
 * shall not be liable for any damages arising out of your use of the sample
 * code, even if IBM has been advised of the possibility of such damages.
 */

package com.ibm.ctg.samples.requestexit;

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.logging.*;

import com.ibm.jzos.FileFactory;

/*
 * 
 * Class: EMInfoLogger
 * 
 * Logging helper class for IBM Supportpac CH0A
 * 
 * Provides synchronized logging output to stdout, z/OS DD destinations,  HFS or UNIX/Windows files
 * 
 * Output will be written to stdout, unless overridden by the system property EM.out
 */

public class EMInfoLogger {

    private static Logger logger;
    private static PrintStream ps;
    private boolean loggingActive = true;
           
    private String destination;     // Log destination
    private String encoding;        // Log encoding
    private int numFiles;           // Number of log files
    private int maxSize;            // Max size of a log file
    private boolean onZos;          // Flag for execution on z/OS operating system

   public EMInfoLogger(String setDestination, String setEncoding, int numOfFiles, int maxFileSize, boolean zosPlatform, String redirectConsole) {
       destination = setDestination;
       encoding = setEncoding;
       numFiles = numOfFiles;
       maxSize = maxFileSize;
       onZos = zosPlatform;       
       Handler handler;  
       
       // Redirect the console (stdout and stderr) to a file if specified.
       if (redirectConsole != null) {
           try {
               File file = new File(redirectConsole);
               FileOutputStream fos = new FileOutputStream(file);
               PrintStream cps = new PrintStream(fos);
               System.setErr(cps);               
               System.setOut(cps);
           } catch (Exception e) {
               printErr("CH0A: Redirect console to '" + redirectConsole + "' failed with " + e);             
           }           
       }
              
       // Now configure the main exit logging
       if (logger == null) {
           // If destination is set, log to file or JES log
           handler = new FlushingStreamHandler(ps,new MessageFormatter());
           if (destination != null)  { // File destination specified
               try {          
                   if (onZos) {
                       encoding = "IBM-1047"; // Force EBCDIC for use in WAS z/OS
                       if (destination.startsWith("//")) { // JES, DD or PDS                                                   
                           // Use com.ibm.jzos.FileFactory to handle DD output if on zOS
                           // Need logic to prevent reopen whilst running
                           BufferedOutputStream bufOut = FileFactory.newBufferedOutputStream(destination);
                           ps = new PrintStream(bufOut);
                           handler = new FlushingStreamHandler(ps,new MessageFormatter(), encoding);
                       } else { // zOS HFS file
                           handler = createFileHandler(destination, numFiles, maxSize, encoding);
                       }
                   } else { // non-zOS system
                       handler = createFileHandler(destination, numFiles, maxSize, encoding);
                   }
               } catch (Exception e) {
                   printErr("CH0A: InfoLogger print to '" + destination + "' failed with " + e);
                   printErr("CH0A: Using default PrintStream");
               }
           } else {
               handler = new ConsoleHandler();
               handler.setFormatter(new MessageFormatter());
           }

           logger = Logger.getLogger("com.ibm.ctg.samples.requestexit.EM");
           // Remove any old handlers
           for(Handler currentHandler : logger.getHandlers()){
               logger.removeHandler(currentHandler);
           }
           logger.setUseParentHandlers(false);
           logger.addHandler(handler);          
       }
   }
   

   /**
    * Creates a Handler to write log messages to a file. If multiple
    * files are required then a FileHandler is used, otherwise our
    * FlushingLogHandler is used.
    * 
    * @param fileName The name of the file to write to
    * @param numFiles The number of log files to be created
    * @param maxSize The maximum size of a log file
    * @param encoding The encoding to use to write log messages
    * @return A handler the logger can use to publish log records
    * @throws IOException If there is an error setting up access to the file 
    */
   private static Handler createFileHandler(String fileName, int numFiles,
         int maxSize, String encoding) throws IOException {
      Handler handler = null;
      if (numFiles > 1) {
         handler = new FileHandler(fileName, maxSize, numFiles, true);
         handler.setFormatter(new MessageFormatter());
         handler.setEncoding(encoding);
      } else {
         handler = new FlushingStreamHandler(new PrintStream(fileName),
               new MessageFormatter(), encoding);
      }
      return handler;
   }
   
   /**
    * Writes single formatted line to the log, including a date/time stamp and identifier
    * 
    * @param msg The body of log message as a String
    */
   public synchronized void writeln(String msg) {
       if (isLoggingActive()) {
           String outputLine;
           String time = formatTime(new Date());
           outputLine=String.format("%s CH0A: %s",time,msg);
           logger.log(Level.INFO,outputLine);           
       }
   }
   
   /**
    * Formats an alert line and writes it to the log, with extended data if enabled.
    * 
    * @param alert The type of alert to be logged
    * @param currentRequest The WorkElement this alert relates to
    * @param logExtendedData Flag indicating if extended data should be logged
    */
   public void logAlert(String alert, ExtendedMonitorWorkElement currentRequest, boolean logExtendedData) {
       
       String msg = new String("");
       
       // Create some local storage for data which might or might not be available
       String server = new String("UNKNOWN");
       String originData = new String("UNKNOWN");
       String xctRoot = new String("UNKNOWN");
       String xctParent = new String("UNKNOWN");
       String xctCurrent = new String("UNKNOWN");
       String startTimeStr = new String("UNKNOWN");
       String endTimeStr = new String("UNKNOWN");

       // Local storage for calculating and storing request times
       long startTime = 0;
       long endTime = 0;
       long reqTime = 0;
       
       startTime = currentRequest.getRequestReceived();
       endTime = currentRequest.getResponseSent();
       startTimeStr = new Date(startTime).toString();
       
       // If we have a response sent time, use it otherwise use the current time 
       if (endTime != 0){
           endTimeStr = new Date(endTime).toString();
           reqTime = endTime - startTime;
       } else {
           reqTime = new Date().getTime() - startTime;
       }
       
       // Report the server selected by DSS if available, otherwise report the server
       // specified by the request. If no server was specified, report 'UNKNOWN' by default
       if (currentRequest.getCicsServer() != null) {
           server = currentRequest.getCicsServer(); 
       } else if (currentRequest.getReqServer() != null) {
           server = currentRequest.getReqServer();
       } 
       
       // Check for and write out extended data if enabled and available
       if (logExtendedData) {
           
           // Check for Origin data availability
           if (currentRequest.getOriginData() != null ) {
               originData = currentRequest.getOriginData().toString();
           } 
           
           // Check for XCT data availability. Assume if XctRoot is available, all XCT data is
           if (currentRequest.getXctRoot() != null) {
               xctRoot = currentRequest.getXctRoot();
               xctParent = currentRequest.getXctParent();
               xctCurrent = currentRequest.getXctCurrent();
           }
           
           // Format a message string with the extended data
           msg = String.format("%s : [Client IP: %s] [Program: %s] [TransId: %s] [Server: %s] [Rc: %s] [Start: %s] [End: %s] [Time: %dms] " +
           		               "[Origin Data: %s] [XCT Root: %s] [XCT Parent: %s] [XCT Current: %s]",
                   alert, currentRequest.getClientLocation().getHostAddress(), currentRequest.getProgram(), currentRequest.getTranName(), 
                   server, currentRequest.getCicsRcString(), startTimeStr, endTimeStr, reqTime, originData, xctRoot, xctParent, xctCurrent);           
       } else {
           // Format a message string without extended data
           msg = String.format("%s : [Client IP: %s] [Program: %s] [TransId: %s] [Server: %s] [Rc: %s] [Start: %s] [End: %s] [Time: %dms] ",
                   alert, currentRequest.getClientLocation().getHostAddress(), currentRequest.getProgram(), currentRequest.getTranName(), 
                   server, currentRequest.getCicsRcString(), startTimeStr, endTimeStr, reqTime);
       }
       // Log the alert using the standard write.
       writeln(msg);
   }
   
   /**
    * Enables log entries to be written
    */
   public synchronized void startLogging() {
       setLoggingActive(true);
   }
   
   /**
    * Stops log entries being written
    */
   public synchronized void stopLogging() {
       setLoggingActive(false); 
   }
   
   /**
    * Sets the logging state to active or inactive based on the flag passed in
    * 
    * @param active
    */
   private synchronized void setLoggingActive(boolean active) {
       synchronized (this) {
           loggingActive = active;
       }      
   }
   
   /**
    * Checks the current state of logging and returns it
    * 
    * @return Boolean indicating if logging is active or inactive
    */
   private boolean isLoggingActive() {
       synchronized (this) {
           return loggingActive;
       }      
   }

   

   /**
    * Write a line to STDERR
    * 
    * @param msg
    */
   private static synchronized void printErr(String msg) {
      System.err.println(msg);
   }

   private static class FlushingStreamHandler extends Handler {
      private Formatter f;
      private OutputStream p;
      private String encoding = System.getProperty("file.encoding");

      public FlushingStreamHandler(PrintStream os, Formatter f) {
         this.p = os;
         this.f = f;
      }

      public FlushingStreamHandler(PrintStream os, Formatter f, String encoding) {
         this.p = os;
         this.f = f;
         this.encoding = encoding;
      }

      @Override
      public synchronized void publish(LogRecord record) {
         String msg;
         byte[] output;
         try {
            msg = f.format(record);
            output = msg.getBytes(encoding);
            p.write(output);
            p.flush();
         } catch (Exception ex) {
            printErr("CH0A: InfoLogger failed in publish() with " + ex);
         }
      }

      @Override
      public void close() throws SecurityException {
         try {
            p.close();
         } catch (IOException e) {
            printErr("CH0A: InfoLogger failed in close() with " + e);
         }

      }

      @Override
      public void flush() {
         try {
            p.flush();
         } catch (IOException e) {
            printErr("CH0A: InfoLogger failed in flush() with " + e);
         }
      }
   }

   private static class MessageFormatter extends Formatter {
      @Override
      public String format(LogRecord record) {
         return record.getMessage() + System.getProperty("line.separator");
      }
   }
   /**
    * Format time, returns formatted time stamp as string
    */
   public String formatTime (Date timestamp)
   {
       SimpleDateFormat dfTime = null;
       dfTime = (SimpleDateFormat) DateFormat.getTimeInstance();
       dfTime.applyPattern("MM/dd/yy HH:mm:ss:SSS");
       String time = dfTime.format(timestamp);
       return time;        
   }
}
