package com.ibm.ctg.samples.requestexit;

/*
 * Licensed Materials - Property of IBM
 * 
 * (C) Copyright IBM Corporation. 2012,2014 All Rights Reserved
 * 
 * US Government Users Restricted Rights - Use, duplication, or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corporation
 * 
 * 
 * The following code is sample code created by IBM Corporation. This sample
 * code is not part of any standard IBM product and is provided to you solely
 * for the purpose of assisting you in the development of your applications. The
 * code is provided 'AS IS', without warranty or condition of any kind. IBM
 * shall not be liable for any damages arising out of your use of the sample
 * code, even if IBM has been advised of the possibility of such damages.
 */

import java.nio.charset.Charset;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import com.ibm.ctg.client.ECIRequest;
import com.ibm.ctg.monitoring.*;

/*
 * 
 * Class: ExtendedMonitor
 * 
 * SupportPac CH0A for IBM CICS Transaction Gateway Version v8.1 or later
 * 
 * Version : 1.1 Date: September 2014
 * 
 * Description: 
 * This SupportPac provides a monitoring utility based on the CICS TG request monitoring exits. 
 * It is designed to assist with system monitoring and problem determination. Output will be
 * written to System.out (stdout) unless overridden by the system property
 * {@link EM.out}
 * z/OS DD names are supported as output destinations when using the CICS TG
 * on z/OS through the use of the jzos class library.  
 *  
 * See SupportPac documentation for further details on how to configure and use
 * this sample
 */
public class ExtendedMonitor implements RequestExit,Runnable {

	// Definitions for system properties for configuration
    public static final String EM_STDOUT_PROPERTY = "EM.out";                      // Log entry destination
    public static final String EM_CONSOLE_PROPERTY = "EM.console";                 // Console destination
	public static final String EM_NUM_FILES_PROPERTY = "EM.log.number";            // Number of log files
	public static final String EM_FILE_SIZE_PROPERTY = "EM.log.size";              // Max log file size
	public static final String EM_LOG_LRTS_PROPERTY = "EM.log.lrts";               // Log long running transactions
    public static final String EM_LOG_REQERRORS_PROPERTY = "EM.log.reqerrors";     // Log requests completing without ECI_NO_ERROR
	public static final String EM_LOG_INTSTATS_PROPERTY = "EM.log.intstats";       // Log interval stats
	public static final String EM_LOG_EXTDATA_PROPERTY = "EM.log.extdata";         // Log extended data (Origin and XCT data)

	public static final String EM_LRT_TIMEOUT_PROPERTY = "EM.lrttimeout";          // Long running transaction timeout
	public static final String EM_INTERVAL_IMEOUT_PROPERTY = "EM.intervaltimeout"; // Interval Timeout
	public static final String EM_REAPER_TIMEOUT_PROPERTY = "EM.reapertimeout";    // Reaper timeout
	public static final String EM_ORPHAN_TIMEOUT_PROPERTY = "EM.orphantimeout";    // Orphan timeout
	
	public static final String EM_VERSION = "1.1.0";
	
	public static final String LRT_ALERT = "!LRT-ALERT!";
	public static final String RQF_ALERT = "!RQF-ALERT!";
	public static final String RPR_ALERT = "!RPR-ALERT!";
	public static final String STK_ALERT = "!STK-ALERT!";
	
    public String destination;         // Log file
	public String encoding;            // JVM encoding
	
	// Logger object which handles output to specified logging destination
	private EMInfoLogger logger; 	
	
    /*
     * The various timeouts which will be used, with suitable defaults.
     * All times are in ms
     */
    private long lrtTimeout = 5000;       /* Long Running Transaction Timeout           */
    private long intervalTimeout = 60000; /* Interval between statistics logging        */
    private long reaperTimeout = 600000;  /* Interval between running reaper processing */
    private long orphanTimeout = 300000;  /* Interval after which requests are reaped   */
    
    /*
     * Flags to control what is logged. By default all logging is enabled but this can 
     * be overridden by Java properties.
     */
    private boolean logLongRunning = true;   /* Log long running transactions                                */
    private boolean logRequestErrors = true; /* Log requests which complete without ECI_NO_ERROR    */
    private boolean logIntervalStats = true; /* Write RequestExit statistics every interval         */
    private boolean logExtendedData = true;  /* Log Origin data and XCT data on alerts if available */
            
    // Internal Counters to track requests going in and out.
    private static long inboundCount = 0;
    private static long outboundCount = 0;

    // A internal FIFO work queue 
    private static BlockingQueue<ExtendedMonitorWorkElement> requestQueue = new LinkedBlockingQueue<ExtendedMonitorWorkElement>();
    
    // In flight requests will be stored in a HashMap - this will be 'reaped' periodically 
    // to prevent abnormal growth. 
    private HashMap<Integer,ExtendedMonitorWorkElement> storedRequests = new HashMap<Integer, ExtendedMonitorWorkElement>();

    /* 
     * Main monitor thread - this thread is the main engine of the exit and is responsible for most
     * of the logging output. It's static so there is one per Gateway, and the statistics 
     * are delivered on a per Gateway basis,.
     */
    private static Thread thrMonitor = null;    

    // Flags to control activity
    private static boolean emError = false;
    private static boolean emShutdown = false;
    private static boolean emPause = false;
    
    // These fields are required to allow the exit to work with CICS TG V8.1 but exploit 
    // the new RequestDetails exit point in CICS TG V9
    private boolean useRequestDetails = false;
    private RequestEvent reqDetails;
   
	/**
	 * Default constructor 
	 * 
	 * This calls the init() methods which will perform any required initialisation.
	 */
	public ExtendedMonitor() {
		init();		
	}

	/**
	 * init()
	 * 
	 * This method performs all the initialisation required by the exit. 
	 */
	public void init () {
	    
	    String destination;      // Log destination
	    String encoding;         // Log encoding
	    int numFiles = 1;        // Number of log files
	    int maxSize = 0;         // Max size of a log file
	    boolean onZos;           // Flag for execution on z/OS operating system
	    String tmpString = null; // Temporary string for reading Java properties
		
	    /*
	     * Before doing anything else, configure the logging engine. If no Java properties are
	     * specified, log messages will go to stdtout.
	     */
	    
	    // Determine if the exit is running on z/OS
	    onZos = System.getProperty("os.name").equals("z/OS"); 
		encoding = Charset.defaultCharset().name();
					
		// Check system properties and override logging settings if required.
		destination = System.getProperty(EM_STDOUT_PROPERTY);
		String maxSizeStr = System.getProperty(EM_FILE_SIZE_PROPERTY);
		if(maxSizeStr != null){
			try{
				maxSize = (Integer.parseInt(maxSizeStr))*1024;
				String numFilesStr = System.getProperty(EM_NUM_FILES_PROPERTY);

				// only enable if not using z/OS //DD file destination
				if( (numFilesStr != null) 
						&& (!destination.startsWith("//"))
						&& (maxSize > 0) ){
					numFiles = Integer.parseInt(numFilesStr);
				}
			} catch(NumberFormatException nfe){
				//No or bad value specified so get out
			}
		}
		
		// Check to see if console output should be redirected to a file 
	    tmpString = System.getProperty(EM_CONSOLE_PROPERTY);
	    String redirectConsole = null;
	    if(tmpString != null){
	        redirectConsole = tmpString;
	    }
		
	      // Initialise the Logger
        logger = new EMInfoLogger(destination, encoding, numFiles, maxSize, onZos, redirectConsole);

		
		/* 
		 * Check the other system properties to determine if the default timeouts need to be modified.
		 * 
		 * Note: Inputs are NOT sanity checked.
		 */
		tmpString = System.getProperty(EM_LRT_TIMEOUT_PROPERTY);
		if (tmpString != null) {
		    lrtTimeout = Long.parseLong(tmpString);
		}
        tmpString = System.getProperty(EM_INTERVAL_IMEOUT_PROPERTY);
        if (tmpString != null) {
            intervalTimeout = Long.parseLong(tmpString);
        }
        tmpString = System.getProperty(EM_REAPER_TIMEOUT_PROPERTY);
        if (tmpString != null) {
            reaperTimeout = Long.parseLong(tmpString);
        }
        tmpString = System.getProperty(EM_ORPHAN_TIMEOUT_PROPERTY);
        if (tmpString != null) {
            orphanTimeout = Long.parseLong(tmpString);
        }
		    	    
	    /*
	     * Configure what to log based on properties. Since everything is logged by default,
	     * this serves to disable some logging if set.
	     * 
	     * Note: Inputs are NOT sanity checked.
	     */
	    tmpString = System.getProperty(EM_LOG_LRTS_PROPERTY);
	    String strOff = "off";
	    if(tmpString != null && (tmpString.equalsIgnoreCase(strOff))){
	        logLongRunning = false;   
	    }
	    
	    tmpString = System.getProperty(EM_LOG_REQERRORS_PROPERTY);
	    if(tmpString != null && (tmpString.equalsIgnoreCase(strOff))){
	            logRequestErrors = false;   
	    }

	    tmpString = System.getProperty(EM_LOG_INTSTATS_PROPERTY);
	    if(tmpString != null && (tmpString.equalsIgnoreCase(strOff))){
	        logIntervalStats = false;
	    }
	    
	    tmpString = System.getProperty(EM_LOG_EXTDATA_PROPERTY);
        if(tmpString != null && (tmpString.equalsIgnoreCase(strOff))){
            logExtendedData = false;
        }
	            
        // Check to see if the RequestDetails event is available (CICS TG V9 and later)
        try {
            if (RequestEvent.valueOf("RequestDetails") != null) {
                useRequestDetails = true;
                reqDetails = RequestEvent.valueOf("RequestDetails");
             }
                
        } catch (Exception e) {
                // This can be empty since the default is to not use RequestDetails
        }
           		
	    // Log a startup message
		String msg=String.format("Starting %s V%s on %s, JVM encoding: %s",
				getClass().getSimpleName(),EM_VERSION,System.getProperty("os.name"),encoding);
		logger.writeln(msg); 
		
		// Log the timeouts being used for clarity
		msg=String.format("[LRT Timeout: %dms] [Interval Timeout: %dms] [Reaper Timeout: %dms] [Orphan Timeout: %dms]",
                           lrtTimeout, intervalTimeout, reaperTimeout, orphanTimeout);
        logger.writeln(msg); 
        
        // Log the logging options for clarity
        msg=String.format("[Log Long Running Transactions: %s] [Log Request Errors: %s] [Log Interval Statistics: %s]",
                          String.valueOf(logLongRunning),String.valueOf(logRequestErrors),String.valueOf(logIntervalStats));
        logger.writeln(msg);         
        
        // Log if the RequestDetails event is available.
        if (useRequestDetails) {
            logger.writeln("Detected CICS TG supports RequestDetails event"); 
        }
        
        
        // Now start the main monitor thread which will do the heavy lifting.
		try {
	        thrMonitor = new Thread(this, "ExtendedMonitorThread");
	        thrMonitor.start();
		} catch (Exception e) {
			emError = true;
			logger.writeln("ExtendedMonitorThread Initialisation Error: Monitoring disabled"); 
			e.printStackTrace();
		}
	}
	
	/*
	 * eventFired()
	 * 
	 *  This method is required for all Request Monitoring exits. CICS TG calls this method with varying
	 *  event types and data, which can then trigger additional processing by the exit. 
	 *  
	 *  Note that for RequestEntry, RequestDetails and ResponseExit event types, this code is executed on the thread
	 *  of execution that is handling a request. This means that any delay here will have a direct effect on the
	 *  response time of a request. To avoid this, request data is placed on a FIFO work queue to be processed
	 *  by the main monitor thread. 
	 */
	public void eventFired(RequestEvent event, Map<RequestData, Object> data) {
	    
	    String msg; // For log messages
	    	    	    
		try {	
 
			//RequestEntry event called when request enters Gateway code
			if ( !emError && !emPause && event.equals(RequestEvent.RequestEntry)) {
			    		 
		        ExtendedMonitorWorkElement inboundRequest = new ExtendedMonitorWorkElement(new Date(),data, event);
		        requestQueue.add(inboundRequest);
			}
			
            // RequestDetails event (CICS TG V9 and later) called after DSS has selected a CICS server.
            else if ( !emError && !emPause && useRequestDetails && event.equals(reqDetails)) {
                
                ExtendedMonitorWorkElement detailRequest = new ExtendedMonitorWorkElement(new Date(),data,event);
                requestQueue.add(detailRequest);
            }

			// ResponseExit event called when request exits Gateway code and returns to client
			else if ( !emError && !emPause && event.equals(RequestEvent.ResponseExit)) {
			    
			    ExtendedMonitorWorkElement outboundRequest = new ExtendedMonitorWorkElement(new Date(),data,event);
		        requestQueue.add(outboundRequest);
			}
						
			/*
			 * Any other events are administration events and should be handled.
			 */			

			// ShutDown event called at Gateway shutdown or exit termination, so allow for graceful end.	
			else if (event.equals(RequestEvent.ShutDown)) {
			    msg = String.format("Shutdown initiated in %s",getClass().getSimpleName());
				logger.writeln(msg);
				emShutdown=true;
			} 

			
			// Handle any commands sent to the Request Monitoring infrastructure
			else if (event.equals(RequestEvent.Command)) {

			    // Log what command has been received.
				String command = (String) data.get(RequestData.CommandData);
                msg = String.format("Command [%s] received", command);
                logger.writeln(msg);

                // If logging is not active, activate it.
				if (command.equalsIgnoreCase("CH0A_START")) {
				    logger.startLogging();
				    emPause = false;
	                msg = String.format("Command [%s] received", command);
	                logger.writeln(msg);
				}
				
				// If logging is active, deactivate it.
				else if (command.equalsIgnoreCase("CH0A_STOP")) {
				    emPause = true;
				    logger.stopLogging();
				}
				
				// Scan the request table for any which breach the Long Running Timeout
				// criteria and log a line for each one.
                else if (command.equalsIgnoreCase("CH0A_LRTDUMP")) {
                    logCurrentLRTs();
                }

				// Any unrecognised commands are logged. This is not necessarily a problem since commands
				// could be intended for other exits. 
				else { 
	                msg = String.format("Command [%s] not recognised. Command ignored.", command);
	                logger.writeln(msg);
				}
			}

			// Unknown event fired - this is a catch all failure case. 
			else {
                logger.writeln("Event not recognised. Event ignored.");
			}

		} catch (NoSuchFieldError e){ 
			//This means that a command was entered and Gateway is at pre-v7.1.0.2 level
			e.printStackTrace();

		}
		catch (Exception e) {
			e.printStackTrace();

		}
	}
	
	
    /*
     * The run() method implements the main Monitor thread. This does all the work
     * so the threads passing through the exit don't need to. The idea is to preserve
     * performance of requests as much as possible.
     *
     */
    public void run() {

        // We need some local storage for various checks and stats
        long inboundMarker = 0;
        long outboundMarker = 0;
        boolean logWritten = false;

        // Transaction counts for a given interval
        long curInTPM = 0;
        long curOutTPM = 0;
        long curInflight = 0;
 
        // Time counts for a given interval
        long cumReqTime = 0;
        long avgTime = 0;
        long hiWater = 0;

        // Various time related objects
        long elapsedTime = 0;
        long intervalTime = 0;
        long reaperTime = 0;
        Date intervalTimeCheck = new Date();
        Date reaperTimeCheck = new Date();
        int curDay = 0;
        
        // Strings for message formating
        String msg;

        try {

            ExtendedMonitorWorkElement workRequest = new ExtendedMonitorWorkElement();

            
            // Loop continuously until shutdown is indicated.
            while (!emShutdown) {
                
                // If there is an work request on the queue, process it
                if (!requestQueue.isEmpty()) {
                    workRequest = requestQueue.take();
                    
                    // Reset the flag to show this request has not been logged yet
                    logWritten = false;

                    // If this is an new inbound request, add it to the request table
                    if (workRequest.getDataType().equals(RequestEvent.RequestEntry)) {
                        // Increment the global count 
                        inboundCount++;

                        // Store the request
                        storedRequests.put(workRequest.getCtgCorrelator(),workRequest);

                    // If this is a RequestDetails event, find the existing request in the request table 
                    // and update appropriate data. This is only possible when running in CICS TG V9 
                    // and later.
                    } else if (workRequest.getDataType().equals(reqDetails)) {
                        
                        ExtendedMonitorWorkElement matchedRequest = storedRequests.get(workRequest.getCtgCorrelator());
                        
                        // If the request was found in the request table, update it 
                        if (matchedRequest != null) {
                            
                            // Update the DSS selected CICS Server
                            if (workRequest.getCicsServer() != null) {
                                matchedRequest.setCicsServer(workRequest.getCicsServer());                                
                            }
                            
                            // Update the Origin data
                            if (workRequest.getOriginData() != null) {
                                matchedRequest.setOriginData(workRequest.getOriginData());
                            }
                            storedRequests.put(matchedRequest.getCtgCorrelator(),matchedRequest);
                        }
                                                                 
                    // If this is an outbound request, find the request in the request table and check to see how long it took 
                    } else if (workRequest.getDataType().equals(RequestEvent.ResponseExit)) {
                        
                        ExtendedMonitorWorkElement matchedRequest = storedRequests.get(workRequest.getCtgCorrelator());
                        
                        // If the request was found in the table, now do the time check and remove the entry from the table 
                        if (matchedRequest != null) {

                            // Increment the global count
                            outboundCount++;

                            // Work out how long this request took
                            elapsedTime = workRequest.getDataTime().getTime() - matchedRequest.getDataTime().getTime();
                            cumReqTime = cumReqTime + elapsedTime;

                            //  Update the high water mark if necessary
                            if (elapsedTime > hiWater) {
                                hiWater = elapsedTime;
                            }
                            
                            // If logging requests which fail, check the return code and log if appropriate.
                            if (logRequestErrors && workRequest.getCicsReturnCode() != ECIRequest.ECI_NO_ERROR) {
                                
                                if ((workRequest.getFlowType().equals(FlowType.EciSynconreturn)) || 
                                            (workRequest.getFlowType().equals(FlowType.ExtendedModeEci)) ||   
                                            (workRequest.getFlowType().equals(FlowType.XaEci))) {
                                    
                                    logger.logAlert(RQF_ALERT, workRequest, logExtendedData);
                                    logWritten = true;
                                            
                                }
                            } 
                            
                            // If logging long running requests, check how long the request took and log if appropriate.
                            // If a request has previously been logged because it failed, it will not be logged again.
                            if (logLongRunning && !logWritten) {
                                
                                // If the request took more than the LRT Trigger time - log an alert 
                                if ( elapsedTime > lrtTimeout ) {

                                    // Different content will be logged depending on the type of request                                                                  
                                    
                                    if ((workRequest.getFlowType().equals(FlowType.EciSynconreturn)) || 
                                        (workRequest.getFlowType().equals(FlowType.ExtendedModeEci)) ||   
                                        (workRequest.getFlowType().equals(FlowType.XaEci))) {
                                        
                                        logger.logAlert(LRT_ALERT, workRequest, logExtendedData);
                                        logWritten = true;
                                    } 
                                }
                            }
                            storedRequests.remove(workRequest.getCtgCorrelator());

                        // A stored request matching this outbound request was not found
                        } else {
                            logger.writeln("Outbound request ignored - matching inbound request not found");
                        }

                        // This model is extensible - other types of work could be added - so a simple catch
                        // to trap anything done by error
                    } else {
                        logger.writeln("Request ignored- unknown request type");                        
                    }
                }

                // This part of the thread is responsible for the output of the regular statistics.
                // This is done periodically, and not every time through the loop
                intervalTime = new Date().getTime() - intervalTimeCheck.getTime();
                if (intervalTime >= intervalTimeout) {

                    // Calculate the various stats. A marker is used to keep track of the global counts 
                    // as last processed so they never need to be reset and can be used for cumulative data 
                    curInTPM = inboundCount - inboundMarker;
                    inboundMarker = inboundCount;
                    curOutTPM = outboundCount - outboundMarker;
                    outboundMarker = outboundCount;
                    curInflight = inboundCount - outboundCount;
                    if (curOutTPM > 0) {
                        avgTime = cumReqTime / curOutTPM;
                    }

                    // It may be possible for this to go negative if close requests are sent from Gateway, 
                    // resulting in more complete requests than sent. In this case we can't know the number
                    // of in-flight requests so we'll set it to 0
                    if (curInflight < 0) {
                        curInflight = 0;
                    }
                    
                    if (logIntervalStats) {
                        // Log the statistics data
                        msg=String.format("!INT-STATS! : [In: %d] [Out: %d] [Inflight: %d] [Total: %d] [Avg Time: %d] [Max Time: %d]",
                                          curInTPM, curOutTPM, curInflight, outboundCount,  avgTime, hiWater);
                        logger.writeln(msg);                         
                    }
                   
                    // Reset those counts used for interval statistics
                    curInTPM = 0;
                    curOutTPM = 0;
                    curInflight = 0;
                    intervalTime = 0;
                    hiWater = 0;
                    avgTime = 0;
                    cumReqTime = 0;

                    // Check the time and date 
                    intervalTimeCheck = new Date();
                    Calendar timeCheck = new GregorianCalendar();
                    timeCheck.setTime(intervalTimeCheck);

                    // According to Java documentation, the last minute of the day is 23:59, so if we are in
                    // that minute, reset the counters for a new day
                    if (curDay == 0) {
                        curDay = timeCheck.get(Calendar.DAY_OF_YEAR);
                    }
                    if (timeCheck.get(Calendar.DAY_OF_YEAR) != curDay) {
                        logger.writeln("End of Day - resetting counts");

                        // Reset the cummulative counters
                        inboundCount = 0;
                        outboundCount = 0;
                        inboundMarker = 0;
                        outboundMarker = 0;
                                           

                        // Empty the table of stored requests
                        storedRequests.clear();
                        requestQueue.clear();

                        // reset the current day to the new one
                        curDay = timeCheck.get(Calendar.DAY_OF_YEAR);
                    }

                }

                /*
                 * Reaper processing needs to be included to cater for any dropped or broken connections. In those cases
                 * when work completes, the exit is not driven, which may leave orphaned objects in the current request table, 
                 * causing it to expand over time. This processing periodically cleans out old entries in the table to 
                 * prevent that. This is only done periodically, and not everytime through the loop - it uses a much longer 
                 * interval than the stats loop
                 */
                reaperTime = new Date().getTime() - reaperTimeCheck.getTime();
                if (reaperTime >= reaperTimeout) {

                    // Loop through the Hashmap, check the time of each object against current time
                    Set<Integer> resultsSet = storedRequests.keySet();
                    Iterator<Integer> reqIterator = resultsSet.iterator();
                    while (reqIterator.hasNext()) {
                        Integer currentCorrelator = reqIterator.next();
                        ExtendedMonitorWorkElement reaperRequest = new ExtendedMonitorWorkElement();
                        reaperRequest = storedRequests.get(currentCorrelator);
                        Date reqTime = reaperRequest.getDataTime();
                        elapsedTime = new Date().getTime() - reqTime.getTime();

                        // Requests taking longer than the specified timeout are probably orphaned,
                        // so reap them and log them
                        if ( elapsedTime > orphanTimeout ) {

                            // If the request is an ECI or XA Request, we can extract some useful data to log, otherwise just warn
                            // that a long running transaction occurred.
                            if ((reaperRequest.getFlowType().equals(FlowType.EciSynconreturn)) || 
                                (reaperRequest.getFlowType().equals(FlowType.ExtendedModeEci)) ||   
                                (reaperRequest.getFlowType().equals(FlowType.XaEci))) {
                                
                                logger.logAlert(RPR_ALERT, reaperRequest, logExtendedData);

                            } else {  
                                msg=String.format("!RPR-ALERT! : [Long running non-CICS request removed from table after %dms]", elapsedTime);
                                logger.writeln(msg); 
                            }

                            reqIterator.remove();
                            outboundCount++;  
                        }
                    }
                    reaperTime = 0;
                    reaperTimeCheck = new Date();
                }

                // If the work queue is empty, the thread may spin rapidly which may use system resources. In this
                // case the thread can sleep for a short time while waiting for work.
                if (requestQueue.isEmpty()) {
                    try {
                        Thread.sleep(1000);
                    } catch (Exception e  ) {
                        emError = true;
                        logger.writeln("!ERROR! Monitor Thread Sleep Error: Monitoring disabled");
                        return;
                    }
                }
            }

        } catch (Exception e  ) {
            emError = true;
            logger.writeln("!ERROR! [Monitor Thread General Error: Monitoring disabled]");
            e.printStackTrace();
            return;
        }

    }
    
    
    /*
     * This method iterates through the table of current requests, and logs any which have
     * been in flight longer than the time specified for lrtTimeout. 
     */    
    private void logCurrentLRTs() {
        
        // Create local variables
        long elapsedTime = 0;
        long currentTime = 0;
        Date reqTime;

        // Get the current time
        currentTime = new Date().getTime();
        
        // Loop through the table, check the time of each object against current time
        Set<Integer> resultsSet = storedRequests.keySet();
        Iterator<Integer> reqIterator = resultsSet.iterator();
        while (reqIterator.hasNext()) {
            Integer currentCorrelator = reqIterator.next();
            ExtendedMonitorWorkElement currentRequest = new ExtendedMonitorWorkElement();
            currentRequest = storedRequests.get(currentCorrelator);
            reqTime = currentRequest.getDataTime();
            elapsedTime = currentTime - reqTime.getTime();

            // Requests taking longer than the specified lrtTimeout are considered long running
            // and possibly stuck so log them.
            if ( elapsedTime > lrtTimeout ) {

                // If the request is an ECI or XA Request, we can extract some useful data to log, 
                // otherwise we will ignore it for the purposes of this exit.
                if ((currentRequest.getFlowType().equals(FlowType.EciSynconreturn)) || 
                    (currentRequest.getFlowType().equals(FlowType.ExtendedModeEci)) ||   
                    (currentRequest.getFlowType().equals(FlowType.XaEci))) {
                    
                    logger.logAlert(STK_ALERT, currentRequest, logExtendedData);
                } 
            }
        }   
    }
}
