package com.ibm.ctg.samples.requestexit;
/*
 * Licensed Materials - Property of IBM
 * 
 * (C) Copyright IBM Corporation. 2012,2014 All Rights Reserved
 * 
 * US Government Users Restricted Rights - Use, duplication, or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corporation
 * 
 * 
 * The following code is sample code created by IBM Corporation. This sample
 * code is not part of any standard IBM product and is provided to you solely
 * for the purpose of assisting you in the development of your applications. The
 * code is provided 'AS IS', without warranty or condition of any kind. IBM
 * shall not be liable for any damages arising out of your use of the sample
 * code, even if IBM has been advised of the possibility of such damages.
 */

import java.net.InetAddress;
import java.util.*;

import com.ibm.ctg.client.ECIRequest;
import com.ibm.ctg.monitoring.*;
import com.ibm.ctg.util.OriginData;

/*
 * 
 * Class: ExtendedMonitorWorkElment
 * 
 * Supportpac CH0A for IBM CICS Transaction Gateway Version v8.0 or later
 * 
 *   Description   : This class encapsulates a work element for the ExtendedMonitor 
 *                   request monitoring exit 
 *
 * See SupportPac documentation for further details on how to configure and use
 * this sample
 */
public class ExtendedMonitorWorkElement  {

    // These fields all map directly to fields provided by the 
    // Request Monitoring infrastructure with the exception of
    // dataTime and dataType, which are created when the exit
    // is called for this request.
    private Date            dataTime;
    private RequestEvent    dataType;
    private int             ctgCorrelator;
    private String          cicsAbendCode;
    private int             cicsReturnCode;
    private String          cicsRcString;
    private String          cicsServer;
    private int             ctgReturnCode;
    private FlowType        flowType;
    private String          program;
    private long            requestReceived;
    private long            requestSent;
    private long            responseReceived; 
    private long            responseSent;
    private int             retryCount;
    private String          tranName;
    private InetAddress     clientLocation;
    private OriginData      originData;
    private String          reqServer;
    private String          xctRoot;
    private String          xctParent;
    private String          xctCurrent;
    
    
    // These fields are required to allow the exit to work with CICS TG V8.1 but exploit 
    // the new RequestDetails exit point and XCT data available in CICS TG V9
    private boolean useRequestDetails = false;
    private RequestEvent reqDetails;
    private boolean useXctDetails = false;
    private RequestData reqXctRoot;
    private RequestData reqXctParent;
    private RequestData reqXctCurrent;
    

    // Simple Constructor to initialise the work element
    public ExtendedMonitorWorkElement () {
        dataTime = null;
        dataType = null;
        ctgCorrelator = 0;       
        cicsAbendCode = null;
        cicsReturnCode = 0;
        cicsRcString = null;
        cicsServer = null;
        ctgReturnCode = 0;
        flowType = null;
        program = null;
        requestReceived = 0;
        requestSent = 0;
        responseReceived = 0; 
        responseSent = 0;
        retryCount = 0;
        tranName = null;
        clientLocation = null;
        originData = null;
        reqServer = null;
        xctRoot = null;
        xctParent = null;
        xctCurrent = null;
        
        
        /*
         * In order to allow the exit to work in CICS TG V8.1, but exploit new data provided
         * in CICS TG V9 or later if running in that environment, the following checks determine 
         * what data points are available at runtime. If the data points are available, mappings
         * are set up manually since definitions introduced at CICS TG V9 or later cannot be 
         * compiled in because they won't work with the CICS TG V8.1 classes.
         */
        try {
            // If RequestDetails is available, the exit is running under CICS TG V9 or later
            // and can exploit additional features
            if (RequestEvent.valueOf("RequestDetails") != null) {
                useRequestDetails = true;
                reqDetails = RequestEvent.valueOf("RequestDetails");
             }
            
            // If XctRoot is defined, assume that XctParent and Current will also be available
            if (RequestData.valueOf("XctRoot") != null) {
                useXctDetails = true;
                reqXctRoot = RequestData.valueOf("XctRoot");
                reqXctParent = RequestData.valueOf("XctParent");
                reqXctCurrent = RequestData.valueOf("XctCurrent");
            }
                
        } catch (Exception e) {
                // This can be empty since we are defaulting to not supporting
        }
   }

    /*
     *  This constructor extracts the relevant data from the map passed in to the exit
     *  by CICS TG and populates an exit work element.
     */
    public ExtendedMonitorWorkElement (Date timestamp, Map<RequestData, Object> data, RequestEvent event) {
        
        // Call default constructor - this is required since the runtime detection of 
        // available data points is in the default constructor.
        this();

        dataTime = timestamp;
        dataType = event;
        ctgCorrelator = (Integer) data.get(RequestData.CtgCorrelator);  
        flowType = (FlowType) data.get(RequestData.FlowType);
        requestReceived = (Long) data.get(RequestData.RequestReceived);
        clientLocation = (InetAddress) data.get(RequestData.ClientLocation);
        reqServer = (String) data.get(RequestData.Server);
        
        // Some data is only available for SYNCONRETURN, Extended ECI calls, and XaEci calls
        if (flowType.equals(FlowType.EciSynconreturn) || flowType.equals(FlowType.ExtendedModeEci)
                || flowType.equals(FlowType.XaEci)) {
            program = (String) data.get(RequestData.Program);
            tranName = (String) data.get(RequestData.TranName);
            
            // Assume if XctRoot is set, the other XCT fields have been too
            if (useXctDetails && data.get(reqXctRoot) != null) {
                xctRoot = (String) data.get(reqXctRoot);
                xctParent = (String) data.get(reqXctParent);
                xctCurrent = (String) data.get(reqXctCurrent);               
            }            
        }
        
        // If the event is RequestDetails, the DSS selected CICS server and Origin data is available 
        if (useRequestDetails && dataType.equals(reqDetails)) {

            if (data.get(RequestData.CicsServer) != null) { 
                cicsServer = (String) data.get(RequestData.CicsServer);
            }
            
            if (flowType.equals(FlowType.EciSynconreturn) || flowType.equals(FlowType.ExtendedModeEci) ||
                flowType.equals(FlowType.XaEci)) {
                
                if (data.get(RequestData.OriginData) != null) {
                    originData = (OriginData) data.get(RequestData.OriginData);
                }
            }
        }
               
        // The following data is only available at ResponseExit
        if (dataType.equals(RequestEvent.ResponseExit)) {
            
            // CICS Abend code and Retry count are only available for requests which actually do work
            if (flowType.equals(FlowType.EciSynconreturn) || flowType.equals(FlowType.ExtendedModeEci) || 
                    flowType.equals(FlowType.XaEci)) {
                
                cicsAbendCode = (String) data.get(RequestData.CicsAbendCode);
                if (data.get(RequestData.RetryCount) != null) {
                    retryCount = (Integer) data.get(RequestData.RetryCount);                    
                }                
            }
            
            if (flowType.equals(FlowType.EciSynconreturn) || flowType.equals(FlowType.ExtendedModeEci) ||
                flowType.equals(FlowType.XaEci)) {
                if (data.get(RequestData.OriginData) != null) {
                    originData = (OriginData) data.get(RequestData.OriginData);
                }
            }
            
            // The following pieces of data are only available for certain calls
            if (flowType.equals(FlowType.EciSynconreturn) || flowType.equals(FlowType.ExtendedModeEci) ||
                    flowType.equals(FlowType.ExtendedModeCommit) || flowType.equals(FlowType.ExtendedModeRollback) ||
                    flowType.equals(FlowType.XaEci)) {
                cicsReturnCode = (Integer) data.get(RequestData.CicsReturnCode);
                ECIRequest tmpReq = new ECIRequest();
                tmpReq.Cics_Rc = cicsReturnCode;
                cicsRcString = tmpReq.getCicsRcString();
                cicsServer = (String) data.get(RequestData.CicsServer);
                originData = (OriginData) data.get(RequestData.OriginData);
            }
                      
            // Timings will only be available where requests actually go to CICS, so check before getting.
            if (data.get(RequestData.RequestSent) != null) {
                requestSent = (Long) data.get(RequestData.RequestSent);
            }
            if  (data.get(RequestData.ResponseReceived) != null) {  
                responseReceived = (Long) data.get(RequestData.ResponseReceived);
            }
            
            // The following data should always be available
            ctgReturnCode = (Integer) data.get(RequestData.CtgReturnCode);
            responseSent = (Long) data.get(RequestData.ResponseSent);   
        } 
     }
    
    // Getters and Setters follow

    public Date getDataTime() {
        return dataTime;
    }

    public RequestEvent getDataType() {
        return dataType;
    }

    public int getCtgCorrelator() {
        return ctgCorrelator;
    }

    public String getCicsAbendCode() {
        return cicsAbendCode;
    }

    public int getCicsReturnCode() {
        return cicsReturnCode;
    }
    
    public String getCicsRcString() {
        return cicsRcString;
    }

    public String getCicsServer() {
        return cicsServer;
    }
    
    public void setCicsServer(String dssServer) {
        cicsServer = dssServer;
    }

    public int getCtgReturnCode() {
        return ctgReturnCode;
    }

    public String getProgram() {
        return program;
    }

    public long getRequestReceived() {
        return requestReceived;
    }

    public long getRequestSent() {
        return requestSent;
    }

    public long getResponseReceived() {
        return responseReceived;
    }

    public long getResponseSent() {
        return responseSent;
    }

    public int getRetryCount() {
        return retryCount;
    }
    
    public String getTranName() {
        return tranName;
    }

    public InetAddress getClientLocation() {
        return clientLocation;
    }

    public OriginData getOriginData() {
        return originData;
    }
    public void setOriginData(OriginData reqOriginData) {
        originData = reqOriginData;
    }      

    public FlowType getFlowType() {
        return flowType;
    }     
    
    public String getReqServer() {
        return reqServer;
    }
    
    public String getXctRoot() {
        return xctRoot;
    }
    
    public String getXctParent() {
        return xctParent;
    }
    
    public String getXctCurrent() {
        return xctCurrent;
    }
}
